-- file YGREC8/VHDL/FSM/FSM_rtl.vhdl
-- created ven. janv. 17 01:54:46 CET 2020 by whygee@f-cpu.org
-- version dim. janv. 19 00:27:12 CET 2020
-- version dim. janv. 19 22:43:05 CET 2020 : verbose generic added
--
-- Released under the GNU AGPLv3 license


Library ieee;
    use ieee.std_logic_1164.all;
Library SLV;
    use SLV.SLV_utils.all;
Library YGREC8;
    use YGREC8.ygrec8_def.all;
Library work;
    use work.all;

entity Y8_FSM is
  generic (
    verbose : integer := 0
  );
  port(
    Clk,             -- nominal-speed clock input
    Reset,           -- The Master Reset pin of the chip. It's not used much elsewhere in the core.
    CmdRdy,          -- Indicate a new command is received.
    SND_PC,          -- '1' when SND field = PC ("111", AND3) AND opcode /= SET|CALL
    Opc_LDCx,        -- '1' when opcode = LDCL or LDCH (AND3)
    Opc_INV,         -- '1' when opcode = INV ("1111") (AND4)
    PC_ov : in SL;   -- PC overflow bit also signals completion of code load
    Command: in tControlCommand; -- 2-bits input
    Status : out tControlState;  -- main FSM
    Phase  : out tInstrPhase; -- sub-FSM
    Decode : out SL);
end Y8_FSM;

architecture behavioural of Y8_FSM is
  signal ControlCommand : tControlCommand;
  signal ControlState : tControlState;
  signal InstrPhase : tInstrPhase;
  signal NewCommand, PrevCmdRdy, cache_SND_PC :  SL;
begin

  Status <= ControlState;
  Phase  <= InstrPhase;

  dbg: if verbose /= 0 generate
    process(ControlCommand, ControlState, InstrPhase) is
    begin
      report "  Command=" & CommandString(ControlCommand)
           & "  State="   & StatusString(ControlState)
           & "  Phase="   & PhaseString(InstrPhase);
    end process;
  end generate;

  process(Clk, Reset) is
    variable RunStep : SL;
  begin
    if Reset = '0' then
      -- just clear the FSMs:
      ControlCommand <= C_START;
      ControlState <= S_RESET;
      InstrPhase <= P_IDLE;
      NewCommand <= '0';
      PrevCmdRdy <= '0';
      if verbose /= 0 then
        report "RESET";
      end if;
    else
      if rising_edge(CLK) then
        -- latch the command
        NewCommand <= '0';
        if PrevCmdRdy = '0' and CmdRdy = '1' then
          ControlCommand <= Command;
          NewCommand <= '1';
        end if;
        -- save for next cycle
        PrevCmdRdy <= CmdRdy;

        -- The System's FSM:
        if ControlCommand = C_RESET then
          ControlState <= S_RESET;
        else
          case ControlState is

            when S_RESET =>
              if ControlCommand /= C_RESET then
                ControlState <= S_LOAD;
              end if;

            when S_LOAD =>
              if PC_ov = '1' then
                ControlState <= S_STOP;
                -- This /trick/ is required to allow the state
                -- to change during STOP, so the core can RUN after RESET.
                NewCommand <= '1';
                if verbose /= 0 then
                  report "LOAD ended";
                end if;
              end if;

            when S_STOP =>
              if NewCommand = '1' then
                if ControlCommand = C_START then
                  ControlState <= S_RUN;
                else
                  if ControlCommand = C_STEP then
                    ControlState <= S_STEP;
                  end if;
                end if;
              end if;

            when S_STEP =>
              ControlState <= S_STOP;

            when S_RUN =>
              if ControlCommand = C_STOP or (Opc_INV = '1' and InstrPhase = P_INST) then
                -- stop immediately if INV to prevent oscillation
                ControlState <= S_STOP;
              else
                if ControlCommand = C_STEP then
                  ControlState <= S_STEP;
                end if;
              end if;

            when others =>
              assert false report "Invalid control FSM state "
                    & SLV_to_bin(ControlState) severity failure;

          end case;
        end if;

        -- just a OR gate ?
        RunStep := '0';
        if ControlState = S_RUN or ControlState = S_STEP then
          RunStep := '1';
        end if;

        -- The instruction scheduling FSM:
        case InstrPhase is

          when P_IDLE =>
            if RunStep = '1' and Opc_INV = '0' then
              InstrPhase <= P_INST;
            end if;

          when P_INST =>
            if Opc_INV = '1' or RunStep = '0' then
              InstrPhase <= P_IDLE;
            else
              if Opc_LDCx = '1' then
                InstrPhase <= P_LDCX;
                cache_SND_PC <= SND_PC;
              else
                if SND_PC = '1' then
                  InstrPhase <= P_WrPC;
                end if;
              end if;
            end if;

          when P_LDCX =>
            if cache_SND_PC = '1' then
              InstrPhase <= P_WrPC;
            else
              if RunStep = '0' then
                InstrPhase <= P_IDLE;
              else
                InstrPhase <= P_INST;
              end if;
            end if;

          when P_WrPC =>
            if RunStep = '0' then
              InstrPhase <= P_IDLE;
            else
              InstrPhase <= P_INST;
            end if;

          when others =>
            assert false report "Invalid phase FSM state "
               & SLV_to_bin(InstrPhase) severity failure;

        end case;
      end if;
    end if;
  end process;

end behavioural;
